
#ifndef _NGXE_H_INCLUDED_
#define _NGXE_H_INCLUDED_

#include <ngxe_ngx.h>
#include <ngxe_debug.h>
#include <ngxe_timer.h>
#include <ngxe_xs.h>
#include <ngxe_bufs.h>
#include <ngxe_connection.h>

#define NXSTART     0x01
#define NXRVBUF     0x02

#define NGXE_START  0x01

#endif /* _NGXE_H_INCLUDED_ */

